#!/bin/bash
#

# build_libtorch.sh

# if this build script complains that it cannot find bzlib.h, 
# you need to install bzip2-devel on Scientific Linux

usage()
{
   echo "USAGE: `basename ${0}` <product_directory> <e17> [<prof>] [tar]"
}

# -------------------------------------------------------------------
# shared boilerplate
# -------------------------------------------------------------------

get_this_dir() 
{
    ( cd / ; /bin/pwd -P ) >/dev/null 2>&1
    if (( $? == 0 )); then
      pwd_P_arg="-P"
    fi
    reldir=`dirname ${0}`
    thisdir=`cd ${reldir} && /bin/pwd ${pwd_P_arg}`
}

get_ssibuildshims()
{
    # make sure we can use the setup alias
    if [ -z ${UPS_DIR} ]
    then
       echo "ERROR: please setup ups"
       exit 1
    fi
    source `${UPS_DIR}/bin/ups setup ${SETUP_UPS}`
    
    setup ssibuildshims ${ssibuildshims_version} -z ${product_dir}
}

set_local_qualifiers()
{
  [[ -n "${basequal}" ]] || ssi_die "Qualifiers mandatory"

  qlist=(`echo ${basequal} | sed -e 's/:/ /g' | tr " " "\n"| sort |tr "\n" " "`)
  for qual in "${qlist[@]}"; do
    case $qual in
      e17)
        cqual=$qual
        ;;
      *)
        ssi_die "Unrecognized qualifier $qual"
    esac
  done
  if [[ -z "$cqual" ]]; then
    ssi_die "Base qualifier is mandatory"
  fi
  extraqual=prof
  basequal=${cqual}
}

set_compiler_flags()
{
  case ${cqual} in
    e1[024579])
      xcc=gcc
      cc=gcc
      ;;
    c[257])
      xcc=clang
      cc=clang
      cmake_opts+=(-DLLVM_ENABLE_LIBCXX=ON -DCLANG_DEFAULT_CXX_STDLIB=libc++)
      ;;
    *)
      echo "ERROR: unrecognized compiler qualifier ${cqual}"
      usage
      exit 1
  esac
}

# -------------------------------------------------------------------
# start processing
# -------------------------------------------------------------------

product_dir=${1}
basequal=${2}
maketar=${3}

if [[ "${maketar:-tar}" != "tar" ]]; then
  echo "WARNING: build type specifier ${maketar} no longer accepted: ignoring." 1>&2
  maketar=${4}
fi

if [ -z ${product_dir} ]
then
   echo "ERROR: please specify the local product directory"
   usage
   exit 1
fi

# -------------------------------------------------------------------
# package name and version
# -------------------------------------------------------------------

package=libtorch
origpkgver=v1_0_1
pkgver=${origpkgver}a
ssibuildshims_version=v1_04_13
pkgdotver=${origpkgver//_/.}
pkgdotver=${pkgdotver#v}
# Downloaded tarball.
pkgtar=${package}-${pkgdotver}
pkgtarfile=${pkgtar}.tar

extra_opts=()

get_this_dir
get_ssibuildshims
set_local_qualifiers

source define_basics NO_DEBUG_SOURCE

if [ "${maketar}" = "tar" ] && [ -d ${pkgdir}/lib ]
then
   ${SSIBUILDSHIMS_DIR}/bin/make_distribution_tarball ${product_dir} ${package} ${pkgver} ${fullqual}
   exit 0
fi

echo "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${flvr})"

# -------------------------------------------------------------------
# package dependent stuff goes here
# -------------------------------------------------------------------

mkdir -p ${pkgdir}
if [ ! -d ${pkgdir} ]
then
   echo "ERROR: failed to create ${pkgdir}"
   exit 1
fi

mkdir -p ${blddir}
if [ ! -d ${blddir} ]
then
   echo "ERROR: failed to create ${blddir}"
   exit 1
fi
# remove existing sourcedir to avoid patch conflicts
rm -rf ${sourcedir}
mkdir -p ${sourcedir}
if [ ! -d ${sourcedir} ]
then
   echo "ERROR: failed to create ${sourcedir}"
   exit 1
fi

# declare now so we can setup
# fake ups declare
fakedb=${product_dir}/${package}/${pkgver}/fakedb
${SSIBUILDSHIMS_DIR}/bin/fake_declare_product ${product_dir} ${package} ${pkgver} ${fullqual}

# setup
setup -B ${package} ${pkgver} -q +${fullqual//:/:+} -z ${fakedb}:${product_dir}:${PRODUCTS} || \
 ssi_die "fake setup failed: unsupported qualifier?"

ncore=`${SSIBUILDSHIMS_DIR}/bin/ncores`

# We need cmake, but we'll take ninja if it's there.
source ${SSIBUILDSHIMS_DIR}/bin/setup_cmake "${product_dir}" || ssi_die "failed to setup cmake"
# Disable ninja build 
ninjaprod=$(latest_version -t ${product_dir} ninja -z "${product_dir}" -z "${PRODUCTS}" 2>/dev/null)

if [[ -n "${ninjaprod}" ]]; then
  eval setup -B ${ninjaprod} -z "${product_dir}" -z "${PRODUCTS}" || \
    ssi_die "unable to set up found product ${ninjaprod}"
  extra_opts+=("-GNinja")
  make=ninja
  makejarg=()
else
  make=make
  makejarg=(-j ${ncore})
fi

# Require python for build
. setup_python || ssi_die "Unable to set up Python for build"

if [ -z ${PYTHONHOME} ]
then
   echo "ERROR: PYTHONHOME is undefined"
   echo "       something is very wrong"
   exit 1
fi

# Print active products.
ups active

python_packages=${blddir}/python-packages
[[ -d "${python_packages}" ]] && rm -rf "${python_packages}"
mkdir "${python_packages}"
pip install -t "${python_packages}" pyyaml typing

export PYTHONPATH=${python_packages}${PYTHONPATH:+:${PYTHONPATH}}

cd ${sourcedir} && rm -rf "${pkgtar}"
tar xf ${tardir}/${pkgtarfile} || ssi_die "Unable to untar source"

set_compiler_flags

set -x

cd ${sourcedir}/${pkgtar}

# Apply source tweaks
rm -f cmake/Modules/FindOpenBLAS.cmake

[[ -n "${DEBUG_PATCHES}" ]] && exit 1

cd ${blddir}

if [ ${xcc} = "gcc" ]
then
echo "Building with gcc: $(which gcc) and g++: $(which g++)" 1>&2
elif [ ${xcc} = "clang" ]
then
echo "Building with clang: $(which clang) and clang++: $(which clang++)" 1>&2
fi

compile_opts+=(CC=${cc}
  REL_WITH_DEB_INFO=1
  MAX_JOBS=${ncore}
  NO_CUDA=1
  NO_CUDNN=1
  NO_FBGEMM=1
  NO_TEST=1
  NO_MIOPEN=1
  NO_MKLDNN=1
  NO_NNPACK=1
  NO_QNNPACK=1
  NO_DISTRIBUTED=1
  NO_SYSTEM_NCCL=1
  USE_OPENCV=0
  USE_FFMPEG=0
  USE_LEVELDB=0
  USE_LMDB=0
  BUILD_BINARY=1)

cmake_opts+=(-DBLAS=OpenBLAS -DBUILD_CUSTOM_PROTOBUF=OFF -DUSE_NCCL=OFF)
compile_opts+=(EXTRA_CAFFE2_CMAKE_FLAGS="${cmake_opts[*]}")

env "${compile_opts[@]}" python \
  ${sourcedir}/${pkgtar}/tools/build_libtorch.py || \
  ssi_die "Unable to build ${package}"

# FIXME Install in the right place.
rsync_cmd=(rsync -lptgoD)
install_tmp=${sourcedir}/${pkgtar}/torch/lib/tmp_install

"${rsync_cmd[@]}" -r "${install_tmp}/" "${pkgdir}/" || \
  ssi_die "Unable to install files to ${pkgdir}/"

set +x

if [ ! -d ${pkgdir}/lib ]
then
   echo "ERROR: failed to create ${pkgdir}/lib"
   echo "       Something is very wrong"
   exit 1
fi

# Fix hard-wired path in one file:
perl -wapi\~ -e \
  'm&^\s*(?i)SET(?-i)\s*\(\s*ATEN_INCLUDE_DIR& and s&/.*tmp_install/include\b&\$ENV{LIBTORCH_INC}&' \
  "${pkgdir}/share/cmake/ATen/ATenConfig.cmake" && \
  rm -f "${pkgdir}/share/cmake/ATen/ATenConfig.cmake~" || \
  ssi_die "Unable to fix path in ATenConfig.cmake"

# real ups declare
${SSIBUILDSHIMS_DIR}/bin/declare_product ${product_dir} ${package} ${pkgver} ${fullqual}

# -------------------------------------------------------------------
# common bottom stuff
# -------------------------------------------------------------------

# this should not complain
echo "Finished building ${package} ${pkgver}"
setup ${package} ${pkgver} -q ${fullqual} -z ${product_dir}:${PRODUCTS} || ssi_die "setup failed"
echo "${package} is installed at ${LIBTORCH_FQ_DIR}"

# this must be last
if [ "${maketar}" = "tar" ] && [ -d ${pkgdir}/lib ]
then
   ${SSIBUILDSHIMS_DIR}/bin/make_distribution_tarball ${product_dir} ${package} ${pkgver} ${fullqual}
fi

exit 0
